/*
    This program implements a producer-consumer type of synchronization where the producers and consumers alternate filling and then emptying the buffer. That is, the producers fill the buffer while the consumers all block. When the buffer is full, the producers all block, the consumers wake up, and the producers remain blocked until the consumers empty the buffer.
*/
#include <windows.h>   // Win32 Sleep()
#include <unistd.h>    // Linux sleep()
#include <stdio.h>
#include "pthread.h"
#include "semaphore.h"

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];

sem_t semaphoreCS;     // critical section semaphore
sem_t semaphoreEmpty;  // bounding semaphores
sem_t semaphoreFull;

int fullSlots = 39;    // counts the number of buffer slots that are full
                       // the buffer will have 77 slots
                       // the buffer will be initialized to be half full

int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   void *putIn(void *vargp);
   void *takeOut(void *vargp);

   pthread_t producerID1;
   pthread_t consumerID1;
   pthread_t producerID2;
   pthread_t consumerID2;
   char producerParam1 = '+';  // thread function parameters
   char consumerParam1 = '0';
   char producerParam2 = 'x';  // thread function parameters
   char consumerParam2 = '@';

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create a binary Semaphore to protect the critical section
   sem_init(&semaphoreCS, 0, 1); // signaled

   // create two counting Semaphores that signal a half full buffer
   sem_init(&semaphoreEmpty, 0, 0);  // not signaled
   sem_init(&semaphoreFull,  0, 39); // signaled

   // create the producer and consumer threads
   printf("Starting two producer and two consumer threads\n");
   pthread_create(&producerID1, NULL, putIn,   &producerParam1);  // should have error checking
   pthread_create(&consumerID1, NULL, takeOut, &consumerParam1);
   pthread_create(&producerID2, NULL, putIn,   &producerParam2);
   pthread_create(&consumerID2, NULL, takeOut, &consumerParam2);

   printf("The producer is thread number %d.\n", producerID1);
   printf("The consumer is thread number %d.\n", consumerID1);
   printf("The producer is thread number %d.\n", producerID2);
   printf("The consumer is thread number %d.\n", consumerID2);

   Sleep(50000);     // let the threads run for a while
   runFlag = FALSE;  // now stop the threads
   Sleep(200);       // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
void *putIn(void *vargp)
{
   int i;
   int r; // will hold a random integer
   // seed the random number generator for this thread
   srand( GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an empty spot in the buffer
      sem_wait(&semaphoreEmpty);

     // begin a critical section
      sem_wait(&semaphoreCS);
      theBuffer[bufferIndex+1] = *((char*)vargp);
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      fullSlots++;
      sem_post(&semaphoreCS);
     // end of critical section

      // Signal the consumers if the buffer is full
      if (fullSlots == 77)  // the buffer is full
         for (i=0; i<77; i++)
            sem_post(&semaphoreFull);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(70);  // sleep a short while
   }
   return 0;
}//putIn


// remove a plus sign from the buffer (the consumer)
void *takeOut(void *vargp)
{
   int i;
   int r;  // will hold a random integer
   // seed the random number generator for this thread
   srand( 2*GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an item in the buffer to remove
      sem_wait(&semaphoreFull);

     // begin a critical section
      sem_wait(&semaphoreCS);
      theBuffer[bufferIndex] = *((char*)vargp);
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      fullSlots--;
      sem_post(&semaphoreCS);
     // end of critical section

      // Signal producers if the buffer is empty
      if (fullSlots == 0)  // the buffer is empty
         for (i=0; i<77; i++)
            sem_post(&semaphoreEmpty);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(120);  // sleep a short while
   }
   return 0;
}//takeOut